#include <cpput/assertstring.h>
#include <algorithm>

namespace CppUT {

CheckerResult 
checkStdStringStartsWith( const std::string &string,
                          const std::string &pattern,
                          const Message &message )
{
   CheckerResult result;
   if ( string.substr( 0, pattern.length() ) != pattern )
   {
      result.setFailed();
      result.message_ = message;
      result.message_.add( translate( "String does not start with the expected pattern." ) );
      result.message_.add( translate( "Expected start: " ) + CppTL::quoteMultiLineString( pattern ) );
      result.message_.add( translate( "Actual string:  " ) + CppTL::quoteMultiLineString( string ) );
   }
   return result;
}


CheckerResult 
checkStdStringEndsWith( const std::string &string,
                        const std::string &pattern,
                        const Message &message )
{
   CheckerResult result;
   if ( string.length() < pattern.length()  ||
        string.substr( string.length() - pattern.length() ) != pattern )
   {
      result.setFailed();
      result.message_ = message;
      result.message_.add( translate( "String does not end with the expected pattern." ) );
      result.message_.add( translate( "Expected end:  " ) + CppTL::quoteMultiLineString( pattern ) );
      result.message_.add( translate( "Actual string: " ) + CppTL::quoteMultiLineString( string ) );
   }
   return result;
}


CheckerResult 
checkStdStringContains( const std::string &string,
                        const std::string &pattern,
                        const Message &message )
{
   CheckerResult result;
   if ( string.find( pattern ) == std::string::npos )
   {
      result.setFailed();
      result.message_ = message;
      result.message_.add( translate( "String does not contain the expected pattern." ) );
      result.message_.add( translate( "Expected pattern: " ) + CppTL::quoteMultiLineString( pattern ) );
      result.message_.add( translate( "Actual string: " ) + CppTL::quoteMultiLineString( string ) );
   }
   return result;
}


CheckerResult 
checkStdStringEquals( const std::string &expected,
                      const std::string &actual,
                      const Message &message )
{
   CheckerResult result;
   if ( actual != expected )
   {
      unsigned int differenceIndex = 0;
      while ( differenceIndex < actual.length()  &&  
              differenceIndex < expected.length()  &&
              actual[differenceIndex] == expected[differenceIndex] )
      {
         ++differenceIndex;
      }

      std::string common = expected.substr( 0, differenceIndex );
      std::string actualTail = actual.substr( differenceIndex );
      std::string expectedTail = expected.substr( differenceIndex );

      result.setFailed();
      result.message_ = message;
      result.message_.add( translate( "Strings are not equal." ) );
      if ( !common.empty() )
         result.message_.add( translate( "Common: " ) + CppTL::quoteMultiLineString( common ) );
      result.message_.add( translate( "Divergence position (0 based): " ) + stringize( differenceIndex ) );
      result.message_.add( translate( "Expected (at divergence): " ) + CppTL::quoteMultiLineString( expectedTail ) );
      result.message_.add( translate( "Actual (at divergence):   " ) + CppTL::quoteMultiLineString( actualTail ) );
   }
   return result;
}

} // namespace CppUT
