#include <cpput/resource.h>

namespace CppUT
{


// //////////////////////////////////////////////////////////////////
// //////////////////////////////////////////////////////////////////
// class ResourceHandlerRegistry
// //////////////////////////////////////////////////////////////////
// //////////////////////////////////////////////////////////////////

   ResourceHandlerRegistry &
   ResourceHandlerRegistry::instance()
   {
      static ResourceHandlerRegistry registry;
      return registry;
   }

   void 
   ResourceHandlerRegistry::addConstant( const ResourceName &name,
                                         const Resource &constant )
   {
      ResourceHandlerPtr handler( new ConstantResourceHandler( constant ) );
      add( name, handler );
   }


   ResourceId 
   ResourceHandlerRegistry::add( const ResourceName &name, 
                                 const ResourceHandlerPtr &handler )
   {
      CppTL::Mutex::ScopedLockGuard guard( lock_ );
      ++nextId_;
      ResourceData data;
      data.name_ = name;
      data.lazyCount_ = 0;
      data.handler_ = handler;
      data.id_ = nextId_;
      resourceById_[nextId_] = data;
      idsByName_[name] = nextId_;
      return nextId_;
   }


   void 
   ResourceHandlerRegistry::remove( ResourceId id )
   {
      CppTL::Mutex::ScopedLockGuard guard( lock_ );

      ResourceById::iterator itResource = resourceById_.find( id );
      if ( itResource == resourceById_.end() )
      {
         return;
      }

      ResourceData &data = itResource->second;
      idsByName_.erase( data.name_ );
      resourceById_.erase( itResource );
   }


   ResourceLazyPtr 
   ResourceHandlerRegistry::getResource( const ResourceName &name )
   {
      CppTL::Mutex::ScopedLockGuard guard( lock_ );
      
      IdsByName::iterator itId = idsByName_.find( name );
      if ( itId == idsByName_.end() )
      {
         return ResourceLazyPtr();
      }

      ResourceId id = itId->second;
      return ResourceLazyPtr( id );
   }


   AcquiredResourceHandlePtr 
   ResourceHandlerRegistry::obtainResource( ResourceId id )
   {
      if ( id == 0 )
      {
         return AcquiredResourceHandlePtr();
      }

      ResourceInstancePtr instance;
      ResourceHandler::AccessPolicy access;
      // Scoped lock to ensure that the retrieval/creation of a resource does
      // prevent the retrieval/creation of another resource.
      // Otherwise, dead-lock would occur if one thread held the lock on a resource
      // while trying to obtain a second one, and the second one would be locked by
      // another thread trying to obtain the first resource.
      {
         CppTL::Mutex::ScopedLockGuard guard( lock_ );
         ResourceById::iterator itData = resourceById_.find( id );
         if ( itData == resourceById_.end() )
         {
            return AcquiredResourceHandlePtr();   // Resource was removed
         }
         ResourceData &data = itData->second;
         access = data.handler_->accessPolicy();
         if ( !data.instance_ )  // Resource not instantiated, create it
         {
            data.instance_.reset( new ResourceInstance( data.name_,
                                                        data.handler_ ) );
         }
         instance = data.instance_;
      }
      // Now that the global lock has been required, we can safely acquire the
      // resource (which may lock until another thread using the resource is done
      // with it if the resource does not support concurrent access).
      return instance->acquire( access );
   }


   void 
   ResourceHandlerRegistry::increaseLazyRef( ResourceId id )
   {
      if ( id == 0 )
      {
         return;
      }

      CppTL::Mutex::ScopedLockGuard guard( lock_ );
      
      ResourceById::iterator itResource = resourceById_.find( id );
      if ( itResource != resourceById_.end() )  // checks that the resource was not removed
      {
         ResourceData &data = itResource->second;
         ++(data.lazyCount_);
      }
   }


   void 
   ResourceHandlerRegistry::decreaseLazyRef( ResourceId id )
   {
      if ( id == 0 )
      {
         return;
      }

      CppTL::Mutex::ScopedLockGuard guard( lock_ );
      
      ResourceById::iterator itResource = resourceById_.find( id );
      if ( itResource != resourceById_.end() )  // checks that the resource was not removed
      {
         ResourceData &data = itResource->second;
         if ( --(data.lazyCount_) == 0 )
         {
            // Last lazy reference discarded, remove the ResourceInstancePtr
            // that was stored to prevent it from being destroyed.
            // Only other existing AcquiredResourceHandlePtr will prevent it from being 
            // destroyed.
            data.instance_.reset();
         }
      }
   }



// //////////////////////////////////////////////////////////////////
// //////////////////////////////////////////////////////////////////
// class ResourceInstance
// //////////////////////////////////////////////////////////////////
// //////////////////////////////////////////////////////////////////


   ResourceInstance::ResourceInstance( const ResourceName &name,
                                       const ResourceHandlerPtr &handler )
      : handler_( handler )
      , name_( name )
   {
      resource_ = handler_->create( name );
   }


   ResourceInstance::~ResourceInstance()
   {
      handler_->destroy( name_, resource_ );
   }


   void 
   ResourceInstance::setUp()
   {
      handler_->setUp( name_, resource_ );
   }


   void 
   ResourceInstance::tearDown()
   {
      handler_->setUp( name_, resource_ );
   }


   AcquiredResourceHandlePtr
   ResourceInstance::acquire( ResourceHandler::AccessPolicy access )
   {
      AcquiredResourceHandlePtr handle( 
         new AcquiredResourceHandle( lock_, this, access ) );
      return handle;
   }


// //////////////////////////////////////////////////////////////////
// //////////////////////////////////////////////////////////////////
// class AcquiredResourceHandle
// //////////////////////////////////////////////////////////////////
// //////////////////////////////////////////////////////////////////

   AcquiredResourceHandle::AcquiredResourceHandle( 
                              CppTL::Mutex &lock,
                              const ResourceInstancePtr &instance,
                              ResourceHandler::AccessPolicy access )
      : lock_( lock )
      , instance_( instance )
      , access_( access )
   {  
      if ( access_ != ResourceHandler::concurrent )
      {
         lock_.lock();
      }
   }

   AcquiredResourceHandle::~AcquiredResourceHandle()
   {
      if ( access_ != ResourceHandler::concurrent )
      {
         lock_.unlock();
      }
   }


   void 
   AcquiredResourceHandle::setUp()
   {
      instance_->setUp();
   }

   
   void 
   AcquiredResourceHandle::tearDown()
   {
      instance_->tearDown();
   }



// //////////////////////////////////////////////////////////////////
// //////////////////////////////////////////////////////////////////
// class ResourceLazyPtr
// //////////////////////////////////////////////////////////////////
// //////////////////////////////////////////////////////////////////


ResourceLazyPtr &
ResourceLazyPtr::operator =( const ResourceLazyPtr &other )
{
   ResourceId oldId = id_;
   id_ = other.id_;
   ResourceHandlerRegistry &registry = ResourceHandlerRegistry::instance();
   registry.increaseLazyRef( id_ );
   registry.decreaseLazyRef( oldId );

   return *this;
}


AcquiredResourceHandlePtr 
ResourceLazyPtr::acquire()
{
   if ( isValid() )
   {
      ResourceHandlerRegistry::instance().obtainResource( id_ );
   }
   return AcquiredResourceHandlePtr();
}


} // namespace CppUT
